#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include <termios.h>
#include "dtr.h"
#include "msleep.h"

static int
DTR_change(int fd, int on)
{
#if defined(TIOCM_DTR) && defined(TIOCMGET) && defined(TIOCMSET)

  int status;
  if (ioctl(fd,TIOCMGET,&status))
    return -1;
  if (on)
    status |= TIOCM_DTR;
  else
    status &= ~TIOCM_DTR;
  if (ioctl(fd,TIOCMSET,&status))
    return -1;

#elif defined(TIOCSDTR)
  struct termios term;

  if (tcgetattr(fd,&term))
    return -1;
  if (on)
    ioctl(fd,TIOCSDTR,&term);
  else
    ioctl(fd,TIOCCDTR,&term);
  if (tcsetattr(fd,TCSADRAIN,&term))
    return -1;

#elif defined(MCGETA) && defined(MDTR)

  long mflags;
  if (ioctl(fd,MCGETA,&mflags))
    return -1;
  if (on)
    mflags |= MDTR;
  else
    mflags &= ~MDTR;
  if (ioctl(fd,MCSETA,&mflags))
    return -1;

#else
#error "no way to set/clear dtr"
#endif

  /* ### sicherheitshalber etwas warten */
  msleep(on ? 200 : 500);
  return 0;
}

int DTR_on(int fd) { return DTR_change(fd,1); }
int DTR_off(int fd) { return DTR_change(fd,0); }
int DTR_off_on(int fd)
{
  if (-1==DTR_change(fd,0))
    return -1;
  msleep(250);
  return DTR_change(fd,1);
}
