#include "divedata.h"
#include "mosquitotools.h"
#include "stralloc.h"
#include "buffer.h"
#include "getln.h"
#include "str.h"
#include "bailout.h"
#include "readwrite.h"
#include "error.h"
#include "scan.h"
#include "caltime.h"

static stralloc rawdata;
static void 
enter(strhash *h,const char *s1, uint32 l1, const char *s2, uint32 l2)
{
  if (!strhash_enter(h,1,s1,l1,1,s2,l2)) oom();
}
#define EFIX(s) (s),str_len(s)

static void
finish_mosquitoget(struct dive *d, void (*callback) (struct dive *))
{
  unsigned int i;
  unsigned long maxdepth=0;
  unsigned long avgdepth=0;
  const char *s;
  unsigned int l;
  const char *r;
  for (i=0;i<d->data.len;i++) {
    if (d->data.ix[i].depth > maxdepth)
      maxdepth=d->data.ix[i].depth;
    avgdepth+=d->data.ix[i].depth;
  }
  d->maxdepth=maxdepth;
  d->totalseconds=d->data.ix[d->data.len-1].endtime;
  d->avgdepth=avgdepth/i;

  /* init dive start time to something not happening normally ... */
  tai_uint(&d->start_time,0);

  s=gv(&d->h_dive,"dive-start-date");
  if (s) {
    const char *seps="--T:";
    unsigned long ul[5];
    for (i=0;i<5;i++) {
      l=scan_ulong(s,&ul[i]);
      if (!l) break;
      if (s[l]!=*seps) break;
      seps++;
      s+=l+1;
    }
    if (i==5) {
      struct caltime ct;
      ct.date.year=ul[0];
      ct.date.month=ul[1];
      ct.date.day=ul[2];
      ct.hour=ul[3];
      ct.minute=ul[4];
      ct.second=0;
      ct.offset=0;
      caltime_tai(&ct,&d->start_time);
    }
  }

  /* fix v0.0 mistakes */
  l=rawdata.len;
  r=rawdata.s;
  s=gv(&d->h_dive,"dive-mode");
  if (!s) {
    if (r && l>3 &&  r[l-4] & 0x40) {
      enter(&d->h_dive,EFIX("dive-mode"),EFIX("free"));
    } else {
      s=gv(&d->h_dive,"oxygen-percentage");
      if (!s) s="0";
      if (str_diff(s,"0"))
	enter(&d->h_dive,EFIX("dive-mode"),EFIX("nitrox"));
      else
	enter(&d->h_dive,EFIX("dive-mode"),EFIX("air"));
    }
  }
  /* XXX */
  callback(d);
}

static unsigned long
find_flags(const char *os, const char *s, unsigned short *o2)
{
  unsigned long ret=0;
  if (!*s) return 0;
#define ISFLAG(flagstr) \
  (str_start(s,flagstr) \
    && (s[(sizeof flagstr)-1]==' ' || !s[(sizeof flagstr)-1]))
  while (*s) {
    if      (ISFLAG("SLOW")) { ret |= DDF_SLOW; }
    else if (ISFLAG("ATTENTION")) { ret |= DDF_ATTENTION; }
    else if (ISFLAG("BOOKMARK")) { ret |= DDF_BOOKMARK; }
    else if (ISFLAG("SURFACED")) { ret |= DDF_SURFACED; }
    else if (ISFLAG("DECO-STOP-NEEDED")) { ret |= DDF_DECO_STOP_NEEDED; }
    else if (ISFLAG("DECO-STOP-VIOLATION")) { ret |= DDF_DECO_STOP_VIOLATION; }
    else if (ISFLAG("SAFETY-STOP-VIOLATION"))
      { ret |= DDF_SAFETY_STOP_VIOLATION; }
    else if (str_start(s,"EAN")) {
      unsigned int l;
      unsigned long ul;
      ret|=DDF_SWITCH;
      l=scan_ulong(s+3,&ul);
      if (!l)
	warning(0,"bad EAN value in dive data: ",os,0,0);
      else {
	*o2=ul;
	if ( ((unsigned long)*o2)!=ul)
	  warning(0,"bad EAN value in dive data: ",os,0,0);
      }
    } else { 
      warning(0,"unknown flag value in dive data: ",os,0,0);
      ret|=DDF_UNKNOWN;
    }
    while (*s && *s!=' ')
      s++;
    while (*s==' ')
      s++;
  }
  return ret;
}

void 
handle_mosquitoget(buffer *b, void (*callback)(struct dive *))
{ 
  stralloc sa=STRALLOC_INIT;
  int is_dive=0;
  static struct dive d;
  reinit_dive(&d,"");

  while (1) {
    const char *s;
    int gotlf;
    sa.len=0;
    if (-1==getln(b,&sa,&gotlf,'\n'))
      xbailout(111,errno,"failed to read dump",0,0,0);
    if (!sa.len) break;
    if (!gotlf)
      xbailout(100,0,"unterminated line in dump",0,0,0);
    sa.s[sa.len-1]=0; /* \n -> \0 */
    if (sa.s[0]==0) continue;
    s=sa.s;
    if (*s=='#') {
      if (d.id0.len && is_dive)
	finish_mosquitoget(&d,callback);
      reinit_dive(&d,s+1);
      is_dive=0;
      continue;
    }
    if (*s=='*') {
      if (!d.id0.len)
	xbailout(111,0,"bad input format: * without leading #",0,0,0);
      if (str_equal(s,"*dive0"))
	is_dive=1;
      continue;
    }
    if (*s==':' && is_dive) {
      unsigned int i;
      i=str_chr(s,' ');
      if (!s[i]) {
	warning(0,"ignoring bad colon line in dive: ",s,0,0);
	continue;
      }
      enter(&d.h_dive,s+1,i-1,EFIX(s+i+1));
      continue;
    }
    if (*s=='R' && is_dive) {
      /* R00 01 02 ...  - raw data. */
      /* needed to work around old versions */
      s++;
      while (*s) {
	unsigned long ul;
	unsigned char c;
	unsigned int l=scan_xlong(s,&ul);
	if (l!=2) xbailout(100,0,"bad raw dive data: ",s,0,0);
	c=ul;
	if (!stralloc_append(&rawdata,&c)) oom();
	s+=l;
	if (*s==' ')
	  s++;
      }
      continue;
    }
    if (*s==' ' && is_dive) {
      static stralloc wa_sa;
      /* work around old bug */
      if (!stralloc_copys(&wa_sa,"D0 0 0")) oom();
      if (!stralloc_cats(&wa_sa,s)) oom();
      if (!stralloc_0(&wa_sa)) oom();
      s=wa_sa.s;
    }
    if (*s=='D' && is_dive) {
      /* Dtime depth-cm depth-foot flags */
      unsigned int i;
      unsigned long ul1;
      unsigned long ul2;
      struct diverecord dr;
      s++;
      i=scan_ulong(s,&ul1);
      if (s[i]!=' ' || !i) {
	warning(0,"ignoring bad data line in dive: ",sa.s,0,0);
	continue;
      }
      s+=i+1;
      i=scan_ulong(s,&ul2);
      if (s[i]!=' ' || !i) {
	warning(0,"ignoring bad data line in dive: ",sa.s,0,0);
	continue;
      }
      s+=i+1;
      /* depth in foot as hex value. skip it. */
      while(*s && *s!=' ')
	s++;
      if (*s==' ') {
	s++;
	dr.flags=find_flags(sa.s,s,&dr.o2_percentage);
      } else
	dr.flags=0;
      dr.endtime=ul1;
      dr.depth=ul2;
      if (!dralloc_append(&d.data,&dr)) oom();
    }
  } 
  if (d.data.len)
    finish_mosquitoget(&d,callback);
}
