/* public domain rewrite */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "error.h"
#include "tai.h"
#include "leapsecs.h"
#include "open.h"
#include "alloc.h"
#include "close.h"
#include "readwrite.h"

struct tai *leapsecs = 0;
int leapsecs_num = 0;

static int init_done = 0;

int 
leapsecs_read(void)
{
  int fd;
  struct stat st;
  struct tai *t;
  int n;
  int i;
  struct tai u;

  fd = open_read("/etc/leapsecs.dat");
  if (fd == -1) {
    if (errno != error_noent) return -1;
    if (leapsecs) alloc_free((char *)leapsecs);
    leapsecs = 0;
    leapsecs_num = 0;
    return 0;
  }

  if (fstat(fd,&st) == -1) { close(fd); return -1; }

  t = (struct tai *) alloc(st.st_size);
  if (!t) { close(fd); return -1; }

  n = read(fd,(char *) t,st.st_size);
  close(fd);
  if (n != st.st_size) { alloc_free((char *)t); return -1; }

  n /= sizeof(struct tai);

  for (i = 0;i < n;++i) {
    tai_unpack((char *) &t[i],&u);
    t[i] = u;
  }

  if (leapsecs) alloc_free((char *)leapsecs);

  leapsecs = t;
  leapsecs_num = n;
  return n;
}

int leapsecs_init(void)
{
  if (!init_done) {
    if (-1 == leapsecs_read())
      return -1;
    init_done = 1;
  }
  return 0;
}
