#include "buffer.h"
#include "getln.h"
#include "bailout.h"
#include "error.h"
#include "str.h"
#include "fmt.h"
#include "open.h"
#include "buffer.h"
#include "fmt.h"
#include "math.h"
#include "out.h"
#include "uogetopt.h"
#include "stralloc.h"
#include "divedata.h"
#include "scan.h"
#include "mosquitotools.h"
#include "attributes.h"
#include "zhl16.h"
#include "scan__float.h"
#include "caltime.h"
#include <unistd.h>
#include <stdio.h>

#define UNLOAD_INTERVAL (15*60)
#define UNLOAD_TOLERANCE (0.01)
#define NEAR_ZERO(x) (fabs(x)<0.0001)

static float  feetfactor=1;

static int need_load;
static int o_load=0;
static int o_feet=0;
static int o_fahrenheit=0;
static unsigned long o_zhl16_desat_factor=0;
static unsigned long o_zhl16_sat_factor=0;
static int o_log_mode=0;
static unsigned long o_contload=0;
static unsigned long o_unload=0;
static float o_ceilings=0;
static unsigned long o_vectors=0;
static float         o_air_pressure=1.0;
static const char *o_zhl16_setting="zhl16c";
static float         o_nitrox_fraction=0;
static int           o_color=1;
static int           o_ndc=0;
static int           o_merge=0;
static struct tai    lastdiveendtime;

static const char *divecolor;

static float baseload;
static unsigned int baseload_set;
static float o_maxwidth=16;
static float o_maxheight=8;
static int more_than_one=0;

#define POS(x,y) do { \
  outs("("); \
  outf((x),4); \
  outs(","); \
  outf((y),4); \
  outs(")"); \
} while(0)

#define SC(from,to,val,space,vz) \
  ((vz)*( (( ((float)val)-(from))/((to)-(from)))  *(space)))
#define X(val) SC(0,maxseconds,val,o_maxwidth,1)
#define YDEPTH(val) SC(0,maxdepth/100.0,val/100.0,o_maxheight,-1)
#define YDELTA(val) SC(mindelta,maxdelta,val,o_maxheight,-1)
#define YLOAD(val) SC(maxload,baseload,val,o_maxheight,-1)
#define YNDC(val) SC(min_nullzeit,(float)max_nullzeit,val,o_maxheight,-1)
#define YMDLOAD(val) SC(max_perc_load_for_depth,0,val,o_maxheight,-1)
#define YMSLOAD(val) SC(max_perc_load_for_surface,0,val,o_maxheight,-1)

static void docolor(const char *s, const char *col)
{
  if (!o_color) return;
  outs(" ");
  outs(s); outs(" \""); outs(col); outs("\" ");
}

static const char *flagcolor(unsigned int flags)
{
  if (flags & DDF_SAFETY_STOP_VIOLATION) return "red4";
  if (flags & DDF_DECO_STOP_VIOLATION) return "red";
  if (flags & DDF_SLOW) return "magenta";
  if (flags & DDF_DECO_STOP_NEEDED) return "blue";
  if (flags & DDF_SURFACED) return "brown";
  if (flags & DDF_ATTENTION) return "purple";
  if (flags & DDF_BOOKMARK) return "green4";
  if (flags & DDF_UNKNOWN) return "green1";
  if (flags & DDF_SWITCH) return "orange";
  return 0;
}

static unsigned long maxdepth;
static unsigned long maxseconds;
static unsigned long maxintervals;
static unsigned long maxunloadseconds;
static float maxdelta;
static float mindelta;
static float maxload;
static float max_perc_load_for_depth=0;
static float max_perc_load_for_surface=0;
static unsigned long min_nullzeit;
static unsigned long max_nullzeit;

static void log_num(const char *s,float fact,int st)
{
  unsigned int l;
  float f;
  l=scan__float(s,&f);
  if (!l) {outs("?"); return;}
  outf(f*fact,st);
}
static void log_temp(const char *s)
{
  unsigned long ul;
  unsigned int l;
  if (!s) s="";
  l=scan_ulong(s,&ul);
  if (!l || s[l]) {
    outs(s); outs(" (?)");
  } else if (o_fahrenheit) {
    outu(ul*1.8+32);
    outs(" (F)");
  } else {
    outs(s); outs(" (C)");
  }
}
static void
log_header(struct dive *d)
{
  strhash *h=&d->h_dive;
  const char *s;
  unsigned int i;
  unsigned long volume=0;
  unsigned long p0=0;
  unsigned long p1=0;
  unsigned int l;
  outs(".TS\ntab(@) expand;\n");
  outs("|c s s s s s|\n");
  outs("|l r|l r|l r|\n");
  outs("|l r|l r|l r|\n");
  outs("|l r|l r|l r|\n");
  outs("|l r|l r|l r|\n");
  outs("|l r|l r|l r|\n");
  outs("|l r|l r|l r|\n");
  outs("|l l s s s s|\n");
  outs(".\n_\n");
  outs(d->id0.s); outs("\n_\n");
  outs("Start @ ");
    s=gv(h,"dive-start-date");
    if (!s) s="unknown";
    outs(s);
  outs("@ Max. Depth ");
    if (o_feet) outs("(f)"); else outs("(m)"); outs("@");
    s=gv(h,"max-depth");
    if (!s) outs("unknown");
    else log_num(s,1/feetfactor,1);
  outs("@ Dive # @");
    s=gv(h,"total-dive-number");
    if (!s) s="";
    outs(s);
    outs(" ");
    outs("\n");

  outs("End @ ");
  {
    unsigned long ul[2];
    struct caltime ct;
    caltime_utc(&ct,&d->start_time,0,0);
    ul[1]=ct.minute+d->totalseconds/60;
    ul[0]=ct.hour+ul[1]/60;
    ul[1]%=60;
    ul[0]%=24;
    outu(ul[0]);
    outs(":");
    if (ul[1]<10)
      outs("0");
    outu(ul[1]);
  }
  outs("@ Avg. Depth ");
    if (o_feet) outs("(f)"); else outs("(m)"); outs("@");
    if (d->avgdepth)
      outf(d->avgdepth/100.0/feetfactor,1);
  outs("@ Dive group # @");
    s=gv(h,"dive-no");
    if (!s) s="";
    outs(s);
    outs("\n");

  outs("Dive time @ ");
  {
    unsigned long ul[2];
    ul[1]=(d->totalseconds/60)%60;
    ul[0]=d->totalseconds/3600;
    outu(ul[0]);
    outs(":");
    if (ul[1]<10)
      outs("0");
    outu(ul[1]);
  }
  outs("@ Surface temperature @");
    s=gv(h,"temperature-at-begin");
    log_temp(s);
  outs("@ Mode @");
    s=gv(h,"dive-mode");
    if (!s) s="";
    if (str_equal(s,"nitrox")) {
      outs("EAN");
      s=gv(h,"oxygen-percentage");
      if (s)
	outs(s);
      else {
	outs("??");
	warning(0,"not oxygen-percentage in dive ",d->id0.s,0,0);
      }
    } else {
      outs(s);
    }
    outs("\n");

  outs("Surface interval @ ");
    s=gv(h,"surface-interval-hours");
    if (!s) s="??";
    outs(s);
    outs(":");
    s=gv(h,"surface-interval-minutes");
    if (!s) s="??";
    outs(s);
  outs("@ Temp. at max. depth @");
    s=gv(h,"temperature-at-max-depth");
    log_temp(s);
  outs("@ Altitude setting @");
    s=gv(h,"altitude-setting");
    if (!s) outs("none"); 
    else {
      unsigned long ul=0;
      l=scan_ulong(s,&ul);
      if (!l) {
	warning(0,"cannot to parse altitude setting ",s," in dive ",d->id0.s);
	outs("??");
      } else {
	switch(ul) {
	case 0: outs("0-300m"); break;
	case 1: outs("300-1500m"); break;
	case 2: outs("1500-3000m"); break;
	}
      }
    }
    outs("\n");

  outs("Bottle volume @ ");
    s=gv(h,"bottle-volume");
    if (!s) s="";
    l=scan_ulong(s,&volume);
    if (!l) volume=0;
    outs(s);
    outs(" liter");
  outs("@ Temperature at end@");
    s=gv(h,"temperature-at-end");
    log_temp(s);
  outs("@ Personal setting @");
    s=gv(h,"personal-setting");
    if (!s) outs("none"); 
    else {
      unsigned long ul=0;
      outs(s);
      l=scan_ulong(s,&ul);
      if (!l) {
	warning(0,"cannot to parse personal setting ",s," in dive ",d->id0.s);
	outs("??");
      } else {
	switch(ul) {
	case 0: outs(" (normal)"); break;
	case 1: outs(" (careful)"); break;
	case 2: outs(" (very careful)"); break;
	}
      }
    }
    outs("\n");

  outs("Pressure at start @ ");
    s=gv(h,"pressure-at-begin");
    if (!s) s="";
    l=scan_ulong(s,&p0);
    if (!l) p0=0;
    outs(s);
  outs("@ Pressure at end@");
    s=gv(h,"pressure-at-end");
    if (!s) s="";
    l=scan_ulong(s,&p1);
    if (!l) p1=0;
    outs(s);
  outs("@ Gas usage per minute @");
    if (volume && p0 && p1 && d->avgdepth) {
      double x;
      double avg_amb_pressure=d->avgdepth/1000.0+1.0;
      x=(volume * (p0-p1)) / ( d->totalseconds/60.0*avg_amb_pressure);
      outf(x,2);
    }
  
    outs("\n");

  outs("_\n");
  outs("Place @ "); 
  s=gv(h,"place"); if (s) outs(s);
  outs("\n");
  outs("_\n");
  outs("Equipment @ "); 
  s=gv(h,"equipment"); if (s) outs(s);
  outs("\n");
  outs(" @ \n"); 
  outs("_\n");
  outs("Buddies @ "); 
  s=gv(h,"buddies"); if (s) outs(s);
  outs("\n");
  outs("_\n");
  for (i=0;;i++) {
    static stralloc sa;
    if (!i)
      outs("Comment ");
    if (!stralloc_copys(&sa,"comment")) oom();
    if (!stralloc_catuint0(&sa,i,0)) oom();
    if (!stralloc_0(&sa)) oom();
    s=gv(h,sa.s);
    if (!s) break;
    outs("@"); outs(s); outs("\n");
  }
  outs("        @\n");
  outs("        @\n");
  outs("_\n");

  outs(".TE\n");
}

static void 
get_gas_percentages(struct dive *d, float *n2_fraction)
{
  const char *s;
  s=gv(&d->h_dive,"oxygen-percentage");
  if (s) {
    unsigned long ul;
    unsigned int l;
    l=scan_ulong(s,&ul);
    if (!l || s[l]) {
      warning(0,"failed to parse oxygen percentage",0,0,0);
      *n2_fraction=0.78;
      return;
    }
    *n2_fraction=(100-ul)/100.0;
    return;
  }
  if (NEAR_ZERO(o_nitrox_fraction))
    *n2_fraction=0.78;
  else
    *n2_fraction=o_nitrox_fraction;
}

static void 
scan_callback(struct dive *d)
{
  unsigned int i;
  float n2_fraction;

  if (d->maxdepth>maxdepth)
    maxdepth=d->maxdepth;
  if (d->totalseconds>maxseconds)
    maxseconds=d->totalseconds;
  if (d->data.len>maxintervals)
    maxintervals=d->data.len;

  get_gas_percentages(d,&n2_fraction);

  if (need_load) {
    unsigned long lastendtime=0;
    float *load;
    unsigned int gewebe;
    float l;

    load=zhl16_alloc(&gewebe);
    if (!load) oom();
    l=zhl16_init_load(load,o_air_pressure,0.78);
    if (!baseload_set++) 
      baseload=l;

    for (i=0;i<d->data.len;i++) {
      unsigned int j;
      zhl16_calc_load(load,o_air_pressure,n2_fraction,
	d->data.ix[i].depth/100.0,
	d->data.ix[i].endtime-lastendtime);
      for (j=0;j<gewebe;j++) {
	if (load[j]>maxload)
	  maxload=load[j];
      }
      lastendtime=d->data.ix[i].endtime;
      if (o_ndc) {
	unsigned long u;
	u=zhl16_calc_no_dec_time(load,o_air_pressure,n2_fraction,
	  d->data.ix[i].depth/100.0);
	if (u<min_nullzeit)
	  min_nullzeit=u;
	if (u>max_nullzeit)
	  max_nullzeit=u;
      }
    }
    if (o_unload) {
      int cont=1;
      for (i=UNLOAD_INTERVAL;cont;i+=UNLOAD_INTERVAL) {
	unsigned int j;
	zhl16_calc_load(load,o_air_pressure,0.78,0,UNLOAD_INTERVAL);
	cont=0;
	for (j=0;j<gewebe;j++) { 
	  if (fabs(load[j]-baseload)>UNLOAD_TOLERANCE)
	    cont|=1;
	}
      }
      maxunloadseconds=i;
    }
    zhl16_free(load);
  }
  if (o_vectors) {
    unsigned long lastdepth=0;
    unsigned long lastendtime=0;
    for (i=0;i<d->data.len;i++) {
      unsigned long de=d->data.ix[i].depth;
      long di;
      if (d->data.ix[i].endtime==0) 
	continue; /* record contains only remarks */
      long interval=d->data.ix[i].endtime-lastendtime;
      di=((long)de)-(long)lastdepth;
      di=di*60/(float)interval;
      if (di>maxdelta)
	maxdelta=di;
      if (di<mindelta)
	mindelta=di;
      lastdepth=de;
      lastendtime=d->data.ix[i].endtime;
    }
  }
}

static strhash h;
static int h_flag;

static void
mark(struct dive *d, const char *text, const char *color, struct diverecord *r)
{	
  unsigned long u;
  static stralloc s;
  static stralloc t;
  char *data;
  uint32 datalen;

  if (!h_flag) {
    if (-1==strhash_create(&h,1,32,strhash_hash)) oom();
    h_flag=1;
  }
  if (!stralloc_copys(&s,d->id0.s)) oom();
  if (!stralloc_cats(&s,"-")) oom();
  if (!stralloc_cats(&s,text)) oom();
  if (!stralloc_0(&s)) oom();

  if (1==strhash_lookup(&h,s.s,s.len,&data,&datalen)) {
    if (!stralloc_copyb(&t,data,datalen-1)) oom(); /* -1: \0 */
    strhash_delete(&h);
    if (!stralloc_cats(&t,", ")) oom();
  } else {
    t.len=0;
    if (o_color && color) {
      if (!stralloc_cats(&t,"\\m[")) oom();
      if (!stralloc_cats(&t,color)) oom();
      if (!stralloc_cats(&t,"] ")) oom();
    }
    if (!stralloc_cats(&t,text)) oom();
    if (o_color && color)
      if (!stralloc_cats(&t,"\\m[]")) oom();
    if (more_than_one) {
      if (!stralloc_cats(&t," in dive ")) oom();
      if (o_color && divecolor) {
	if (!stralloc_cats(&t,"\\m[")) oom();
	if (!stralloc_cats(&t,divecolor)) oom();
	if (!stralloc_cats(&t,"] ")) oom();
      }
      if (!stralloc_cats(&t,d->id0.s)) oom();
    }

    if (o_color && divecolor) {
      if (!stralloc_cats(&t,"\\m[]")) oom();
    }
    if (!stralloc_cats(&t," at ")) oom();
  }
  u=r->endtime;
  if (!stralloc_catuint0(&t,u/3600,0)) oom();
  if (!stralloc_cats(&t,":")) oom();
  u%=3600;
  if (!stralloc_catuint0(&t,u/60,2)) oom();
  if (!stralloc_cats(&t,":")) oom();
  if (!stralloc_catuint0(&t,u%60,2)) oom();
  if (!stralloc_0(&t)) oom();
  if (-1==strhash_enter(&h,1,s.s,s.len,1,t.s,t.len)) oom();
}

static void 
do_annotation(void)
{
  char *k,*d;
  uint32 kl,dl;
  if (h_flag) {
    strhash_walkstart(&h);
    while (strhash_walk(&h,&k,&kl,&d,&dl)) {
      outb(d,dl); 
      outs("\n.br\n");
    }
    h_flag=0;
    strhash_destroy(&h);
  }
}

static void annotate(struct dive *d, struct diverecord *r)
{
  static stralloc s;
  if (!stralloc_copys(&s,"Gas switch (EAN")) oom();
  if (!stralloc_catulong0(&s,r->o2_percentage,0)) oom();
  if (!stralloc_cats(&s,")")) oom();
  if (!stralloc_0(&s)) oom();
#define OF(fv,text) \
    if (r->flags & fv) \
	mark(d,text,flagcolor(fv),r);
  OF(DDF_SLOW,"Slow warning appeared");
  OF(DDF_ATTENTION,"Attention");
  OF(DDF_BOOKMARK,"Bookmark");
  OF(DDF_SURFACED,"Surfaced");
  OF(DDF_DECO_STOP_NEEDED,"Deco stop needed");
  OF(DDF_DECO_STOP_VIOLATION,"Deco stop violated");
  OF(DDF_SAFETY_STOP_VIOLATION,"Mandatory safety stop violated");
  OF(DDF_UNKNOWN,"Unknown mark");
  OF(DDF_SWITCH,s.s);
}


static void 
work_callback(struct dive *d)
{
  unsigned int i;
  static float *load;
  static float *max_load_for_depth;
  static float *max_load_for_surface;
  float *lastload=0; /* keep gcc quiet */
  static unsigned int gewebe;
  unsigned long lastendtime=0;
  float lastdepth=0;
  unsigned int lastinterval;
  float circle_size;
  float big_circle_size;
  float n2_fraction;

  get_gas_percentages(d,&n2_fraction);

  if (!more_than_one) {
    if (o_log_mode)
      log_header(d);
    outs(".PS\n");
    outs("maxpswid=");outf(o_maxwidth/2.54,2);outs("\n");
    outs("maxpsht=");outf(o_maxheight/2.54,2);outs("\n");
  }
  if (!tai_less(&lastdiveendtime,&d->start_time)) {
    if (o_contload)
      warning(0,"disabling continuous mode: bad dive order",0,0,0);
    o_contload=0;
  }

  if (need_load) {
    float *tload; /* keep gcc quiet */
    unsigned int j;
    if (!load) {
      load=zhl16_alloc(&gewebe);
      max_load_for_depth=zhl16_alloc(&gewebe);
      max_load_for_surface=zhl16_alloc(&gewebe);
      zhl16_init_load(load,o_air_pressure,0.78);
      if (!load || !max_load_for_depth || !max_load_for_surface) oom();
    } else if (o_contload) {
      struct tai x;
      tai_sub(&x,&d->start_time,&lastdiveendtime);
      zhl16_calc_load(load,o_air_pressure,0.78,0,tai_approx(&x));
    } else
      zhl16_init_load(load,o_air_pressure,0.78);

    tload=zhl16_alloc(&gewebe);
    if (!tload) oom();
    lastload=zhl16_alloc(&gewebe);
    if (!lastload) oom();
    zhl16_copy(tload,load);
    zhl16_copy(lastload,load);

    for (i=0;i<d->data.len;i++) {
      zhl16_calc_load(tload,o_air_pressure,n2_fraction,
	  d->data.ix[i].depth/100.0,
	  d->data.ix[i].endtime-lastendtime);
      for (j=0;j<gewebe;j++)
	if (tload[j]>maxload)
	  maxload=tload[j];
      zhl16_get_max_load_for_depth(max_load_for_depth,o_air_pressure,
	n2_fraction,d->data.ix[i].depth/100.0);
      zhl16_get_max_load_for_depth(max_load_for_surface,o_air_pressure,
	n2_fraction,0);
      for (j=0;j<gewebe;j++) {
	float x;
	x=(tload[j]-baseload)/(max_load_for_depth[j]-baseload);
	if (x>max_perc_load_for_depth)
	  max_perc_load_for_depth=x;
	x=(tload[j]-baseload)/(max_load_for_surface[j]-baseload);
	if (x>max_perc_load_for_surface)
	  max_perc_load_for_surface=x;
      }
      lastendtime=d->data.ix[i].endtime;
    }
    zhl16_free(tload);
  }
  circle_size=o_maxwidth/d->data.len/4;
  if (circle_size>o_maxwidth/20.0)
    circle_size=o_maxwidth/20.0;
  big_circle_size=circle_size*2;
  if (big_circle_size<0.05) 
    big_circle_size=0.05;

  for (lastendtime=0,i=0;i<d->data.len;i++) {
    float depth=d->data.ix[i].depth/100.0;
    unsigned long endtime=d->data.ix[i].endtime;
    unsigned long flags=d->data.ix[i].flags;
    const char *fc;
    unsigned int g;
    int flag_ceiling_drawn=0;

    fc=flagcolor(flags);

    outs("I"); outu(i); outs(": ");
    outs("circle rad "); outf(fc ? big_circle_size : circle_size,4);
    if (more_than_one) {
      if (fc)
	docolor("shaded",fc);
      if (divecolor)
	docolor("outline",divecolor);
    } else if (fc)
      docolor("color",fc);
    outs(" at ");
    POS(X(endtime),YDEPTH(depth*100));
    outs("\n");

    /* lastendtime==endtime if this record has endtime 0, meaning that
     * it contains some remark only. */
    if (lastendtime!=endtime) {
      outs("line ");
      if (divecolor)	
	docolor("color",divecolor);
      outs("from ");
      POS(X(lastendtime), YDEPTH(lastdepth*100.0));
      outs(" to ");
      POS(X(endtime), YDEPTH(depth*100.0));
      outs("\n");
    }
    annotate(d,&d->data.ix[i]);
    if (d->data.ix[i].flags & DDF_SWITCH) {
      if (NEAR_ZERO(o_nitrox_fraction)) {
	n2_fraction=(100-d->data.ix[i].o2_percentage)/100.0;
      }
    }
    if ((o_load || o_ceilings>0 || o_unload || o_ndc) && endtime) {
      float maxperc=0;
      int lead_tissue=-1;
      zhl16_calc_load(load,o_air_pressure,n2_fraction,depth,
	  d->data.ix[i].endtime-lastendtime);
      zhl16_get_max_load_for_depth(max_load_for_depth,o_air_pressure,
	n2_fraction, depth);
      zhl16_get_max_load_for_depth(max_load_for_surface,o_air_pressure,
	n2_fraction,0);
      for (g=0;g<gewebe;g++) {
	float x;
	x=(load[g]-baseload)/(max_load_for_depth[g]-baseload);
	if (x>maxperc && x>0.1) {
	  maxperc=x;
	  lead_tissue=g;
	}
      }
      if (lead_tissue==-1) {
	for (g=0;g<gewebe;g++) {
	  float x;
	  x=(load[g]-baseload)/(max_load_for_surface[g]-baseload);
	  if (x>maxperc && x>0.1) {
	    maxperc=x;
	    lead_tissue=g;
	  }
	}
      }
      for (g=0;g<gewebe;g++) {
	static stralloc sa;
	if (!stralloc_copys(&sa,"GCOLOR")) oom();
	if (!stralloc_catuint0(&sa,g,0)) oom();
	if (!stralloc_0(&sa)) oom();
	if (o_load==1) {
	  outs("line ");
	  if (-1!=lead_tissue && g==(unsigned int)lead_tissue)
	    docolor("color","red");
	  else
	    docolor("color",sa.s);
	  outs("from ");
	  POS(X(lastendtime), YLOAD(
	    lastload[g] > baseload ? lastload[g] : baseload ));
	  outs(" to ");
	  POS(X(endtime), YLOAD(load[g] > baseload ? load[g] : baseload));
	  outs("\n");
	  lastload[g]=load[g];
	} else if (o_load==2 || o_load==3) { /* load rel. to surface or depth */
	  float x;
	  float vgl;
	  if (o_load==2)
	    vgl=max_load_for_depth[g];
	  else
	    vgl=max_load_for_surface[g];
	    
	  x=(load[g]-baseload)/(vgl-baseload);
	  if (lastendtime) {
	    float xx;
	    outs("line ");
	    docolor("color",sa.s);
	    xx=lastload[g]>0 ? lastload[g] : 0;
	    outs(" from ");
	    if (o_load==2) POS(X(lastendtime), YMDLOAD(xx));
	    else           POS(X(lastendtime), YMSLOAD(xx));
	    outs(" to ");
	    xx=x>0 ? x : 0;
	    if (o_load==2) POS(X(endtime), YMDLOAD(xx));
	    else           POS(X(endtime), YMSLOAD(xx));
	    outs("\n");
	  }
	  lastload[g]=x;
	}
      }

      if (o_ceilings>0) {
	zhl16_get_max_load_for_depth(max_load_for_depth, o_air_pressure,
	  0.78,0);
	for (g=0;g<gewebe;g++)
	  if ((load[g]-baseload)/(max_load_for_depth[g]-baseload)
	      >o_ceilings)
	    break;
	if (g!=gewebe) { /* ceiling exists */
	  unsigned long ul;
	  for (ul=0;flag_ceiling_drawn==0;ul+=25) {
	    zhl16_get_max_load_for_depth(max_load_for_depth, o_air_pressure,
	      n2_fraction,ul/100.0);
	    for (g=0;g<gewebe;g++)
	      if ((load[g]-baseload)/(max_load_for_depth[g]-baseload)
		  >o_ceilings)
		break;
	    if (g==gewebe) {
	      static stralloc sa;
	      outs("box ht "); outf(circle_size*2,4); 
	      if (!stralloc_copys(&sa,"GCOLOR")) oom();
	      if (!stralloc_catuint0(&sa,g,0)) oom();
	      if (!stralloc_0(&sa)) oom();
	      docolor("outline","magenta");
	      outs("wid "); outf(circle_size,4); 
	      outs(" at ");
	      POS(X(endtime),YDEPTH(ul));
	      outs("\n");
	      flag_ceiling_drawn=1;
	    }
	  }
	}
      }
    }
    if (o_vectors && endtime) {
      float delta=depth-lastdepth;
      long interval=endtime-lastendtime;
      delta=delta*60/(float)interval;
      if (fabs(YDELTA(delta)-YDELTA(0))>0.05) {
	outs("line -> ");
	docolor("color","VECTORCOLOR");
	outs("from ");
	POS(X(endtime),YDELTA(0));
	outs(" to ");
	POS(X(endtime),YDELTA(delta));
	outs("\n");
	outflush();
      }
    }
    if (o_ndc && min_nullzeit!=max_nullzeit && !flag_ceiling_drawn) {
      unsigned long u;
      u=zhl16_calc_no_dec_time(load,o_air_pressure,n2_fraction,
	d->data.ix[i].depth/100.0);
      outs("circle rad "); outf(circle_size,4);
      outs(" ");
      docolor("color","NDCCOLOR");
      outs("at ");
      POS(X(endtime),YNDC(u));
      outs("\n");
      outflush();
    }
    lastinterval=endtime-lastendtime;
    lastendtime=endtime;
    lastdepth=depth;
  }
  if (o_unload) {
    unsigned long endtime=d->data.ix[d->data.len-1].endtime;
    unsigned long interval=endtime-d->data.ix[d->data.len-2].endtime;
    int cont=1;
    float *tload;
    unsigned int g;
    tload=zhl16_alloc(&gewebe);
    if (!tload) oom();
    for (g=0;g<gewebe;g++)
      tload[g]=load[g];
    for (i=UNLOAD_INTERVAL;cont;i+=UNLOAD_INTERVAL) {
      zhl16_calc_load(tload,o_air_pressure,0.78,0,UNLOAD_INTERVAL);
      cont=0;
      for (g=0;g<gewebe;g++) { 
	static stralloc sa;
	if (fabs(tload[g]-baseload)>UNLOAD_TOLERANCE)
	  cont|=1;

	if (!stralloc_copys(&sa,"GCOLOR")) oom();
	if (!stralloc_catuint0(&sa,g,0)) oom();
	if (!stralloc_0(&sa)) oom();

	outs("line ");
	docolor("color",sa.s);
	outs("from ");
	POS(X(endtime+(i/UNLOAD_INTERVAL)*interval), YLOAD(lastload[g]));
	outs(" to ");
	POS(X(endtime+(i/UNLOAD_INTERVAL)*interval), YLOAD(tload[g]));
	outs("\n");

	lastload[g]=tload[g];
      }
    }
    zhl16_free(tload);
  }
  if (lastload)
    zhl16_free(lastload);
  {
    struct tai x;
    tai_uint(&x,lastendtime);
    tai_add(&lastdiveendtime,&d->start_time,&x);
  }
}

static void numfunc_f(float x) { outf(x,2); }
static void numfunc_s(float x) { outf(x/feetfactor,2); }
static void numfunc_d(unsigned long x) {outu(x/100.0/feetfactor+0.5); }
static void numfunc_t(unsigned long x) 
{ outu(x/60); outs(":"); x%=60; if (x<10) outs("0"); outu(x); }

#define ONEINFO(min,max,color,turn,right,numfunc,pos,val) \
do { \
 outs("T: box "); \
 if (color) docolor("outline",color); \
 outs(" invis \""); \
 if (turn) \
   numfunc(max-(val*100.0-0)/(maxdepth/feetfactor-0)*(max-min)); \
 else \
   numfunc((val*100.0-0)/(maxdepth/feetfactor-0)*(max-min)+min); \
 outs("\" wid myboxwid with ."); \
 if (right) outs("l"); else outs("r"); \
 outs(" at (toff,"); \
 outf(YDEPTH(pos*100*feetfactor),2);\
 outs(")\n"); \
 if (!right) \
   outs("if T.l.x < nmoff then {nmoff = T.l.x}\n"); \
} while(0)
#define ONETYPE(min,max,color,label,unit,turn,right,numfunc) \
  do { \
    unsigned long border=maxdepth/feetfactor/100.0; \
    for (i=0;i<=border;i++) { \
      if (i%step_y) continue; \
      /* leave it out if this one is to near to the last one */ \
      if (YDEPTH(i*100*feetfactor)<YDEPTH(border*100*feetfactor)+0.33) \
	continue; \
      ONEINFO(min,max,color,turn,right,numfunc,i,i); \
    } \
    ONEINFO(min,max,color,turn,right,numfunc,border,border); \
    outs("box "); \
    if (unit) outs("ht 1 "); \
    if (color) docolor("outline",color); \
    outs("invis \""); outs(label); \
    if (unit) { outs("\" \""); outs(unit); } \
    outs("\" with .n at T.s\n"); \
  } while(0)

static void 
frame(void)
{	
  int step_x,step_y;
  unsigned int i;
  unsigned int minutes;
  outs("LLEFT: line from ");
  POS(X(0),YDEPTH(0));
  outs(" to ");
  POS(X(0),YDEPTH(maxdepth));
  outs("\n");

  outs("LBOTTOM: line from last line.end to ");
  POS(X(maxseconds),YDEPTH(maxdepth));
  outs("\n");
  outs("LRIGHT: line from last line.end to "); 
  POS(X(maxseconds),YDEPTH(0));
  outs("\n");
  outs("LRIGHT: line dotted from last line.end to "); 
  POS(X(0),YDEPTH(0));
  outs("\n");

  minutes=maxseconds/60;
  if (minutes<1)
    step_x=5;
  else if (minutes<3) 
    step_x=10;
  else if (minutes<5)
    step_x=20;
  else {
    step_x=1;
    if (minutes/step_x > 15)
      step_x=2;
    if (minutes/step_x > 15)
      step_x=5;
    if (minutes/step_x > 15)
      step_x=10;
    if (minutes/step_x > 15)
      step_x=30;
    if (minutes/step_x > 15)
      step_x=60;
    step_x*=60;
  }
  for (i=step_x;i<maxseconds;i+=step_x) {
    outs("for x=0 to ");
    outf(YDEPTH(maxdepth),4);
    outs(" by ");
    outf(YDEPTH(maxdepth/50.0),4);
    outs(" do {\n");
    outs("  \".\" at ("); outf(X(i),4); outs(",-x)\n");
    outs("}\n");

    outs("\"\" \"");
    if (step_x>=60)
      outu(i/60);
    else {
      outu(i/60);
      outs(":");
      if (i%60 < 10)
	outs("0");
      outu(i%60);
    }
    outs("\"");
    outs(" with .n at ");
    POS(X(i),YDEPTH(maxdepth));
    outs("\n");
  }


  outs("myboxwid = 1.2\n");
  outs("toff = "); outf(X(maxseconds),2); outs("\n");
  outs("nmoff = 0\n");
  /* find out how many horizontal lines to to draw.
   * rule: keep number of lines below 2 per cm.
   * note: maxdepth is in centimeters.
   * note: maxheight in in millimeters.
   */
  step_y=1; /* every meter */
  if (maxdepth/feetfactor/100.0/step_y > o_maxheight*2)
    step_y=2; /* every 2 meter */
  if (maxdepth/feetfactor/100.0/step_y > o_maxheight*2)
    step_y=5;
  while (maxdepth/feetfactor/100.0/step_y > o_maxheight*2)
    step_y*=2;
  ONETYPE(0,maxdepth,0,"Depth",o_feet ? "Feet" : "Meter",0,1,numfunc_d);

  /* draw horizontal lines */
  for (i=1;i<=maxdepth/feetfactor/100.0;i++) {
    int partially=i%step_y;
    if (partially) {
      if (o_feet && i%5)
	continue;
      outs("line dotted from "); 
      POS(X(0)-0.05,YDEPTH(i*100*feetfactor));
      outs(" to ");
      POS(X(0)+0.05,YDEPTH(i*100*feetfactor));
      outs("\n");
      outs("line dotted from "); 
      POS(X(maxseconds)-0.05,YDEPTH(i*100*feetfactor));
      outs(" to ");
      POS(X(maxseconds)+0.05,YDEPTH(i*100*feetfactor));
      outs("\n");
      continue;
    }

    outs("line dotted from "); 
    POS(X(0)-0.05,YDEPTH(i*100*feetfactor));
    outs(" to ");
    POS(X(maxseconds)+0.05,YDEPTH(i*100*feetfactor));
    outs("\n");
  }


  outs("nmoff = 0\n");
  outs("toff = 0\n");

  if (o_vectors) {
    ONETYPE(mindelta,maxdelta,"VECTORCOLOR","Speed",
      o_feet ? "f/s" : "m/s",0,0,numfunc_s);
    outs("toff=nmoff\n");
  }
  if (o_load==1) {
    ONETYPE(baseload,maxload,"GCOLOR8","Load","Bar",1,0,numfunc_f);
    outs("toff=nmoff\n");
  } else if (o_load==2) {
    ONETYPE(0,max_perc_load_for_depth*100,"GCOLOR8","Load","%Amb",1,0,outu);
    outs("toff=nmoff\n");
  } else if (o_load==3) {
    ONETYPE(0,max_perc_load_for_surface*100,"GCOLOR8","Load","%Sur",1,0,outu);
    outs("toff=nmoff\n");
  }
  if (o_ndc && min_nullzeit!=max_nullzeit) {
    ONETYPE(min_nullzeit,max_nullzeit,"NDCCOLOR","NDC","Min.",0,0,numfunc_t);
    outs("toff=nmoff\n");
  }

  if (o_unload) {
    for (i=UNLOAD_INTERVAL;i<=maxunloadseconds;i+=UNLOAD_INTERVAL) {
      if (i%3600)
	continue;
      outs("box invis \"");
      outu(i/3600);
      outs("\"");
      outs(" with .n at ");
      POS(X(maxseconds+i/UNLOAD_INTERVAL*20),YDEPTH(maxdepth));
      outs("\n");
    }
  }
}

static uogetopt2 opts[]={
  {'1',"merge", uogo_flag,UOGO_NOARG, &o_merge,1,
    "Merge dives into one graph.",
    "The default is to create one graph per dive. Use this option to "
    "create an 'overlay' style graph.",0},
  {'A',"air-pressure-at-surface", do_pressure_opt,0, &o_air_pressure,1,
   "Air pressure at surface.",
   "Append  an  unit  to  the number, unless the unit is the default "
   "unit 'bar'. Default value: 1bar.","PRESSURE"},
  {'b',"black-white", uogo_flag,UOGO_NOARG, &o_color,0,
    "Do not use colors.",
    "Color support depends on relatively new versions of groff. Use this "
    "option if your groff is too old.",0},
  {'c',"continuous-load", uogo_flag,UOGO_NOARG, &o_contload,1,
    "Keep track on tissue load between dives.",
    "The default is to start with freshly initialized load values for each "
    "graph. This options causes the program to keep track of the tissue load "
    "between dives.\n"
    "The order of the dives on the command line is important.",0},
  {'C',"ceilings", uogo_float,0, &o_ceilings,1,
    "Show deco ceilings.",
   /*12345678901234567890123456789012345678901234567890*/
    "Draw markers are the deepest depth where a tissue "
    "load would be NUMBER times 100 percent of it's "
    "maximum load or higher. IOW: To maintain a safety "
    "reserve of (1-NUMBER)*100 percent the diver shall "
    "stay below that border.\n"
    "Example: Draw markers at the depth were the diver would have "
    "cross the 25% safety margin:\n"
    "  mosquitograph -ceilings 0.75 filename.dive","NUMBER"},
  {'D',"desaturation-factor", uogo_float,0, &zhl16_desat_factor,1,
    "Desaturation speed factor.",
    "Assume that desaturation halfvalue times are FACTOR times longer "
    "as usual. The default FACTOR is 1.00. Values over one are safer, "
    "and values below one should be avoided, as they are highly "
    "optimistical.","FACTOR"},
  {  0,"feet", uogo_flag,UOGO_NOARG, &o_feet,1,
    "Use feet, not meters.",
    "For use in backward countries.",0},
  {  0,"fahrenheit", uogo_flag,UOGO_NOARG, &o_fahrenheit,1,
    "Use degree fahrenheit, not degree celsius.",
    "For use in one backward country.",0},
  {'H',"height", uogo_float,0, &o_maxheight,0,
    "Height of graphic in cm.",0,0},
  {'l',"load", uogo_flag,UOGO_NOARG, &o_load,1,
    "Show inert gas load in bar.",0,0},
  {  0,"load-ambient-%", uogo_flag,UOGO_NOARG, &o_load,2,
    "Inert gas load as % of tolerable ambient pressure.",
    "Show the inert gas saturation of the tissues as percentage of the "
    "maximum saturation allowed for the tissue at the current depth. "
    "0% is the inert gas "
    "load at the surface, when breathing air, and 100% is the maximum "
    "inert gas load load allowed by the decompression model used.",0},
  {  0,"load-surface-%", uogo_flag,UOGO_NOARG, &o_load,3,
    "Inert gas load as % of tolerable surface pressure.",
    "Show the inert gas saturation of the tissues as percentage of the "
    "maximum saturation allowed for the tissue at the surface. "
    ,0},
  {'m',"model", uogo_string,0, &o_zhl16_setting,0,
    "Decompression model to use.",
    "Model description files are looked for in "
    "/package/"PACKAGE_CLASS"/"PACKAGE_VERSION"/data/ "
    "or the current directory. The file name shall be "
    "MODEL.model","MODEL"},
  {'n',"nitrox-fraction", uogo_float,0, &o_nitrox_fraction,1,
    "The fraction of nitrogen in the cylinder.",
    "In other words: Nitrogen partial pressure in gas mix at 1 bad "
    "ambient pressure. Default: 0.78.\n"
    "Note: This overrides _all_ gas information in the dive data.",0},
  {'N',"no-deco-times", uogo_flag,UOGO_NOARG, &o_ndc,1,
    "Show no deco times.",
    "'No deco time' is the time the diver may stay at "
    "the current depth without having to do a "
    "decompression stop.\n"
    "Note that this is a very rough approximation and that "
    "it doesn't take into account the settings of the "
    "--ceilings and --desaturation-factor options. So "
    "better do not bet your life on this calculation.",0},
  {'P',"page-for-log", uogo_flag,UOGO_NOARG, &o_log_mode,1,
    "Make log file page.",
    "This will create an additional header containing the basic dive data.",0},
  {'S',"saturation-factor", uogo_float,0, &zhl16_sat_factor,0,
    "Saturation speed factor.",
    "Assume  that saturation halfvalue times are FACTOR times as long "
    "as usual. The default FACTOR is 1.\n"
    "Values larger 1 should be avoided at all cost.\n"
    "See the --desaturation-factor option for how to slow down desaturation.",
    "FACTOR"},
  {'V',"vectors", uogo_flag,UOGO_NOARG, &o_vectors,1,
    "Show ascend/descend speed vectors.",
    "This adds red arrows to the graph showing the vertical speed "
    "of the dive.",0},
  {'W',"width", uogo_float,0, &o_maxwidth,0,"Width of graphic in cm.",
    0,0},
  {'u',"unload", uogo_flag,UOGO_NOARG, &o_unload,1,
    "Show inert gas unloading.",
    "This option is broken right now.",0},
  {0,"examples",uogo_print_help,UOGO_NOARG|UOGO_EXIT|UOGO_NOLHD,0,0,
    "Usage examples.",
    "Simple usage, showing the result in the screen:\n"
    "  mosquitograph filename | groff -pt | gv -\n"
    "Create multiple graphs, printing the result:\n"
    "  mosquitograph filename1 filename2 | groff -pt | lpr\n"
    "Show the tissue load on the terminal:\n"
    "  mosquitograph --load filename | groffer\n"
    "Show tissue load and speed vectors:\n"
    "  mosquitograph --load --vectors filename | groffer\n"
    "Show tissue load relative to the maximum for the current depth:\n"
    "  mosquitograph --load-ambient-% filename | groffer\n"
    ,0},
  {  0,"see-also",uogo_print_help,UOGO_NOLHD|UOGO_NOARG|UOGO_EXIT,
    0,0,
  "Where to find related information.",
  "mosquitoget(1) reads dive profiles from suunto mosquito dive computers.\n"
  "dive-profile(5) describes the dive profile data format.\n"
  "make-dive-profile(1) creates dive profiles.\n"
  COMMON_RELATED_INFO
  ,0},
  {0,"",uogo_include,0,general_opts, 0,0,0,0},
  {0,0,0,0,0,0,0,0,0}
};
static uogetopt_env optenv={
  0,PACKAGE,VERSION,
      "usage: mosquitograph FILE [...]\n",
      "Create graphs from dive profiles.",
"mosquitograph creates PIC graphs from one or more dive profiles.\n"
"By default it will create a graph on an own page per dive.\n"
"'No decompression times', speed vectors and inert gas saturation may be\n"
"included."
"The dive profile will be read from the standard input if FILE is `-'."
      ,
      COMMON_BUGREPORT_INFO,
      2,0,
      0,0,
      uogetopt_out,
      opts
};
static const char *
get_divecolor(unsigned int max, uint32 n)
{
  static stralloc sa;
  uint32 mask;
  if (!n) return 0; /* black */
  if (!o_color) return 0;
  mask=0x1000000/max*n;
  //  mask=(((0x08*n)%0xff)<<16)+(((0x18*n)%0xff)<<8)+(0x28*n)%0xff; /* 16*(n%8); */
  mask=((0x80+n*47)%256<<16)
    +((0x40+n*31)%256<<8)
    +((0x00+n*19)%256);
  outs(".defcolor DCOLOR");
  outu(n);
  outs(" rgb #");
  outx(mask,6);
  outs("\n");

  if (!stralloc_copys(&sa,"DCOLOR")) oom();
  if (!stralloc_catuint0(&sa,n,0)) oom();
  if (!stralloc_0(&sa)) oom();
  return sa.s;
}


static void
do_common(void)
{
  maxdelta/=100.0;
  mindelta/=100.0;
}

int 
main(int argc, char **argv)
{
  unsigned int i;
  uogetopt2 *o;
  flag_bailout_fatal_begin=3;
  bailout_progname(argv[0]);
  o=uogetopt_join(opts);
  if (!o) oom();
  optenv.opts=o;
  optenv.program=flag_bailout_log_name;

  uogetopt_parse (&optenv, &argc, argv);

  zhl16_setting(o_zhl16_setting);
  if (o_color)
    outs(".color\n");
  if (o_zhl16_sat_factor)	
    zhl16_sat_factor=o_zhl16_sat_factor/100.0;
  if (o_zhl16_desat_factor)	
    zhl16_desat_factor=o_zhl16_desat_factor/100.0;
  if (o_feet)
    feetfactor=FOOT;
  if (o_color) {
    /* GCOLOR at least is needed for ceilings, too */
    unsigned int gewebe;
    unsigned int j;
    float start[3]={0xa0,0x40,0xff};
    float end[3]=  {0x00,0xff,0x40};
    float add[3];
    float v[3];

    zhl16_free(zhl16_alloc(&gewebe));
    for (j=0;j<3;j++) {
      add[j]=(end[j]-start[j])/(gewebe-1);
      v[j]=start[j];
    }
    for (i=0;i<gewebe;i++) {
      outs(".defcolor GCOLOR");
      outu(i);
      outs(" rgb #");
      for (j=0;j<3;j++) {
	outx(v[j],2);
	v[j]+=add[j];
      }
      outs("\n");
    }
  }
  if (o_vectors && o_color) {
    outs(".defcolor VECTORCOLOR");
    outs(" rgb #c00060\n");
  }
  if (o_ndc && o_color) {
    outs(".defcolor NDCCOLOR");
    outs(" rgb #c08000\n");
  }
  outs("   \n");
  outs("   \n");
  outs(".br\n");
  outs(".br\n");
  outs("   \n");
  outs("   \n");
  outs(".br\n");
  outs(".br\n");
  if (o_unload || o_load || o_ceilings>0 || o_ndc) 
    need_load=1;
  
  if (!o_merge || !argv[2]) {
    divecolor=get_divecolor(0,0);
    for (i=1;argv[i];i++) {
      maxload=0;
      max_perc_load_for_depth=0;
      max_perc_load_for_surface=0;
      maxdepth=0;
      maxseconds=0;
      maxdelta=0;
      mindelta=0;
      max_nullzeit=0;
      min_nullzeit=MAX_NO_DECO_TIME;
      format_handler(argv[i],scan_callback);
      do_common();
      format_handler(argv[i],work_callback);
      frame();
      outs(".PE\n.br\n.br\n \n");
      do_annotation();
      if (argv[i+1])
	outs(".bp\n");
    }
  } else {
    more_than_one=1;
#define DISABLE(flag,name) do { \
if (flag>0) {flag=0; warning(0,"disabling ",name," in merged dive output",0); }\
}while(0)
    DISABLE(o_contload,"continuous load mode");
    DISABLE(o_load,"load mode");
    DISABLE(o_unload,"unload mode");
    DISABLE(o_vectors,"vector mode");
    DISABLE(o_ceilings,"decompression ceilings");
    DISABLE(o_ndc,"no decompression times");
    DISABLE(o_log_mode,"log headers");
    for (i=1;argv[i];i++)
      format_handler(argv[i],scan_callback);
    outs(".PS\n");
    outs("maxpswid=");outf(o_maxwidth/2.54,2);outs("\n");
    outs("maxpsht=");outf(o_maxheight/2.54,2);outs("\n");
    for (i=1;argv[i];i++) {
      divecolor=get_divecolor(argc-1,i-1);
      format_handler(argv[i],work_callback);
    }
    frame();
    outs(".PE\n.br\n.br\n \n");
    do_annotation();
  }
outs(".psbb\n");
  outflush();
  return 0;
}
