#include "bailout.h"
#include "error.h"
#include "str.h"
#include "math.h"
#include "alloc.h"
#include "open.h"
#include "getln.h"
#include "scan.h"
#include "scan__float.h"
#include "attributes.h"
#include "readwrite.h"
#include "close.h"
#include "zhl16.h"
#include <stdio.h>


#define MAX_GEWEBE 64
static unsigned int gewebe=0;
float zhl16_sat_factor=1.0;
float zhl16_desat_factor=1.0;

static float *halftime;
static float *m0;
static float *dm;
static float *a;
static float *b;
static float baseload; /* internal */

static void bad_model_line(const char *f, const char *l) attribute_noreturn;
static void bad_model_line(const char *f, const char *l)
{ xbailout(100,0,"bad line in model file ",f,": ",l); }

void 
zhl16_setting(const char *name)
{
  int fd;
  stralloc sa=STRALLOC_INIT;
  stralloc ln=STRALLOC_INIT; 
  buffer buf;
  char bspace[BUFFER_INSIZE];

  if (!halftime) {
    halftime=(float *) alloc(sizeof(float) * MAX_GEWEBE);
    m0=(float *) alloc(sizeof(float) * MAX_GEWEBE);
    dm=(float *) alloc(sizeof(float) * MAX_GEWEBE);
    a=(float *) alloc(sizeof(float) * MAX_GEWEBE);
    b=(float *) alloc(sizeof(float) * MAX_GEWEBE);
    if (!halftime || !m0 || !dm || !a || !b) oom();
  }
  gewebe=0;

  if (!stralloc_copys(&sa,"./")) oom();
  if (!stralloc_cats(&sa,name)) oom();
  if (!stralloc_cats(&sa,".model")) oom();
  if (!stralloc_0(&sa)) oom();
  fd=open_read(sa.s);
  if (-1==fd && errno==error_noent) {
    if (!stralloc_copys(&sa,"/package/")) oom();
    if (!stralloc_cats(&sa,PACKAGE_CLASS)) oom();
    if (!stralloc_cats(&sa,"/")) oom();
    if (!stralloc_cats(&sa,PACKAGE_VERSION)) oom();
    if (!stralloc_cats(&sa,"/data/")) oom();
    if (!stralloc_cats(&sa,name)) oom();
    if (!stralloc_cats(&sa,".model")) oom();
    if (!stralloc_0(&sa)) oom();
  }
  fd=open_read(sa.s);
  if (-1==fd)
    xbailout(111,errno,"failed to open ",sa.s,0,0);
  buffer_init(&buf,(buffer_op_t) read, fd, bspace,sizeof(bspace));
  while (1) {
    char *s;
    ln.len=0;
    unsigned long ul;
    unsigned int l;
    int gotlf;
    if (-1==getln(&buf,&ln,&gotlf,'\n'))
      xbailout(111,errno,"failed to read ",sa.s,0,0);
    if (!ln.len) break;
    if (!gotlf)
      xbailout(100,0,"unterminated line in ",sa.s,0,0);
    ln.s[ln.len-1]=0; /* \n */
    if (!ln.s[0]) continue;
    if (ln.s[0]=='#') continue;
    /* 'c' compartiment(U) halftime(float) m(float) dm(float) */
    if (ln.s[0]!='c') continue;
    s=ln.s+1;
    while (*s==' ') s++;

#define BAD() bad_model_line(sa.s,ln.s)
    l=scan_ulong(s,&ul);
    if (!l || !ul) BAD();
    if (ul>gewebe) gewebe=ul;
    s+=l;
    while (*s==' ') s++;

    l=scan__float(s,&halftime[ul-1]);
    if (!l) BAD();
    s+=l;
    while (*s==' ') s++;

    l=scan__float(s,&m0[ul-1]);
    if (!l) BAD();
    s+=l;
    while (*s==' ') s++;
    m0[ul-1]/=10.0;

    l=scan__float(s,&dm[ul-1]);
    if (!l) BAD();
    s+=l;
    if (*s) BAD();
    b[ul-1]=1/dm[ul-1];
    a[ul-1]=m0[ul-1]-dm[ul-1];
  }
  close(fd);
}

#ifndef M_E
# define M_E            2.7182818284590452354   /* e */
#endif

#define RQ 0.8 /* bhlmann: 1.0. Us-Navy: 0.9. Schreiner: 0.8 */
static inline float
Palv(float inert_pressure0, float pressure_in_bar)
{
/* if this formula is changed, than zhl16_get_max_load_for_depth has to
 * be changed, too! */
//  return (pressure_in_bar - PWV37 + (1-RQ)/RQ * PCO2) * inert_pressure0;
  return (pressure_in_bar*inert_pressure0*(1-PWV37));
}

void 
zhl16_get_max_load_for_depth(float *maxloads, float air_pressure, 
  float nitrox_anteil, float depth)
{
  unsigned int i;
  float inert_pp;
  float pressure;
  pressure=depth/10+air_pressure;
  inert_pp=Palv(nitrox_anteil, pressure);
  for (i=0;i<gewebe;i++)
//    maxloads[i]=inert_pp / b[i] + a[i];
// Folgerung aus deep ocean:
    maxloads[i]=(depth/10+air_pressure)/b[i]+a[i];
}

float 
zhl16_min_save_depth(float *load, float air_pressure, float nitrox_anteil)
{
  unsigned int g;
  float m=0;
  for (g=0;g<gewebe;g++) {
    float x;
#if 0
    x=(load[g]-a[g])*b[g];
#else
    x=load[g]*b[g]-a[g];
#endif
    x/=nitrox_anteil*(1-PWV37);
    x-=air_pressure;
    if (x>m)
      m=x;
  }
  return m*10;
}

float *zhl16_alloc(unsigned int *gew)
{
  char *x;
  if (!gewebe) xbailout(100,0,"internal error: gewebe == 0",0,0,0);
  x=alloc(sizeof(float)*gewebe);
  *gew=gewebe;
  return (float *) x;
}
void zhl16_free(float *x) {alloc_free((char *)x); }
float 
zhl16_init_load(float *x, float air_pressure, float nitrox_anteil)
{
  unsigned int i;

  baseload=Palv(nitrox_anteil, air_pressure);
  for (i=0;i<gewebe;i++)
    x[i]=baseload;
  return baseload;
}
void 
zhl16_copy(float *to, float *from)
{
  unsigned int i;
  for (i=0;i<gewebe;i++)
    to[i]=from[i];
}


void
zhl16_calc_load(float *load, float air_pressure, float nitrox_anteil,
  float depth, unsigned int interval)
{
  float pressure;
  float minutes;
  unsigned int i;
  float inert_pp;

  pressure=depth/10+air_pressure;
  minutes=interval/60.0;
  inert_pp=Palv(nitrox_anteil, pressure);
#if 0
  if (inert_pp<baseload )
    xbailout(100,0,"internal error: inert_pp < baseload",0,0,0);
#endif
  for (i=0;i<gewebe;i++) {
    float ex;
    float ht;
    if (inert_pp<load[i])
      ht=halftime[i]*zhl16_desat_factor;
    else
      ht=halftime[i]*zhl16_sat_factor;

    ex=-1 * minutes / ht * log(2);
    load[i]=load[i]+(inert_pp - load[i])*(1-exp(ex));
#if 0
    if (pressure>1 && load[i]<baseload)
{
fprintf(stderr,"ie: load[%u]=%.2f, base=%.2f, minu=%.2f, diff=%.2f\n",
i,load[i],baseload, minutes,(inert_pp - load[i])*(1-exp(ex)));
      xbailout(100,0,"internal error: load < baseload",0,0,0);
}
#endif
  }
}

unsigned long 
zhl16_calc_no_dec_time(float *load, float surface_pressure, 
  float inert_pressure_in_resp_gas, float depth)
{
  unsigned int g;
  float t0min = MAX_NO_DECO_TIME;
  unsigned int t0ming = gewebe;
  float inert_pp;
  float pressure;
  pressure=depth/10+surface_pressure;
  inert_pp=Palv(inert_pressure_in_resp_gas,pressure);
  for(g = 0; g < gewebe; g++) {
    float te;
    float Palv0;
    float ml0; /* max load at surface */
    Palv0=Palv(inert_pressure_in_resp_gas,surface_pressure);
    ml0=Palv0/b[g]+a[g];

    if (load[g]>ml0)
/* if ( (load[g]-Palv0)/(ml0-Palv0) > 0.75) */
      te=0; /* must not dive to surface */
      else if (inert_pp<ml0)
/*    else if (inert_pp<(ml0-Palv0)*0.75)  */
      /* will never reach critical point */
      te=MAX_NO_DECO_TIME;
    else if (inert_pp<load[g])
      /* this tissue loses gas */
      te=MAX_NO_DECO_TIME;
    else {
      float ht;
      if (inert_pp<load[g])
	ht=halftime[g]*zhl16_desat_factor;
      else
	ht=halftime[g]*zhl16_sat_factor;
      te=-ht*log( (inert_pp-ml0)/(inert_pp-load[g]))/log(2)*60;
#if 0
      te=-ht*log( (inert_pp-Palv0-(ml0-Palv0)*0.75)/
	(inert_pp-load[g]))/log(2)*60;
#endif
    }
    if(te < t0min) {
      t0min = te;
      t0ming = g;
    }
  }
  return t0min > 0 ? t0min:0;
} 

